from pybrary.ascii import clean_name

from setux.logger import silent
from setux.core.action import Action


class Downloader(Action):
    '''Download a file

    Params:
        url  = source
        dest = destination path
        name = action name
        sudo = user
        mode = destination mode
    '''

    @property
    def label(self):
        return f'<- {self.name}'

    def check(self):
        dest = self.target.file.fetch(self.dest, sudo=self.sudo)
        return dest.size and dest.size > 0

    def deploy(self):
        if self.sudo:
            bak = self.dest
            self.target.dir('/tmp/setux/dload', mode=777, sudo=self.sudo, verbose=False)
            self.dest = f'/tmp/setux/dload/{clean_name(self.name)}'

        ok = self.target.download(
            url = self.url,
            dst = self.dest,
        )

        if self.sudo:
            ret, out, err = self.target.run(f'cp {self.dest} {bak}', sudo=self.sudo, report='quiet')
            ok = ok and ret==0
            self.dest = bak

        self.target.file(self.dest, verbose=False,
            user  = self.user,
            group = self.group,
            mode  = self.mode,
            sudo  = self.sudo,
        )

        return ok


class Sender(Action):
    '''Upload a file'''

    @property
    def labeler(self):
        return silent

    @property
    def label(self):
        dst = f' -> {self.dst}' if self.dst != self.src else ''
        return f'send {self.src}{dst}'

    def check(self):
        lhash = self.local.file(self.src, verbose=False).hash
        rhash = self.target.file(self.dst, sudo=self.sudo, verbose=False).hash
        return rhash == lhash

    def deploy(self):
        if hasattr(self, 'sudo') and self.sudo:
            return self.target.do_send_as(self.sudo, self.src, self.dst)
        else:
            return self.target.do_send(self.src, self.dst)


class Syncer(Action):
    '''Sync a directory'''

    @property
    def label(self):
        dst = self.dst if hasattr(self, 'dst') else self.src
        dst = f' -> {dst}' if dst != self.src else ''
        return f'sync {self.src}{dst}'

    def check(self):
        dst = self.dst if hasattr(self, 'dst') else self.src
        lhash = self.local.dir(self.src, verbose=False).hash
        rhash = self.target.dir(dst, verbose=False).hash
        return rhash == lhash

    def deploy(self):
        dst = self.dst if hasattr(self, 'dst') else self.src
        return self.target.do_sync(self.src, dst)



from setux.core.action import Actions, Action
from setux.actions.user import User_
from setux.actions.transfer import Sender


class Sudoer(Action):
    '''Add User to sudoers.

    context:
        user : user name
    '''

    @property
    def label(self):
        return f'Sudoer {self.user}'

    def check(self):
        grp = self.target.groups.fetch(self.user)
        ok = 'wheel' in grp.get()

        ret, out, err = self.target.run(f'sudo -l -U {self.user}')
        ok = ok and '(ALL) NOPASSWD: ALL' in (line.strip() for line in out)
        return ok

    def deploy(self):
        grp = self.target.groups.fetch(self.user)
        grp.add('wheel')

        ok = self.target.write(
            f'/etc/sudoers.d/{self.user}',
            f'{self.user} ALL=(ALL) NOPASSWD: ALL',
        )
        return ok


class CopyId(Action):
    '''Send Public Key to Target.

    context:
        user : User name
        pub  : Public key
    '''

    @property
    def label(self):
        return f'Copy ID {self.user}'

    def check(self):
        if not getattr(self, 'pub', None): return True

        user = self.target.user.fetch(self.user)
        if user.check() is not True: return False

        path = f'/home/{self.user}/.ssh/authorized_keys'
        pub = self.target.file.fetch(
            path, mode='600', user=self.user, group=user.group.name
        )
        ok = pub.check() is True
        if ok:
            ok = pub.hash == self.local.file(self.pub, verbose=False).hash

        return ok

    def deploy(self):
        user = self.target.user.fetch(self.user)

        path = f'/home/{self.user}/.ssh'
        ssh = self.target.dir(
            path, mode='700', user=self.user, group=user.group.name
        )
        if ssh.check() is not True: return False

        full = f'{path}/authorized_keys'
        self.target.run('rm -f {full}', sudo='root')
        sent = Sender(self.target, src=self.pub, dst=full, **self.context)()
        if sent is not True: return False

        key = self.target.file(
            full, mode='600', user=self.user, group=user.group.name
        )
        return key.check() is True


class KnownHosts(Action):
    '''Send known_hosts to Target.

    context:
        user : User name
    '''

    @property
    def label(self):
        return 'Send known_hosts'

    def check(self):
        user = self.target.user.fetch(self.user)
        if user.check() is not True: return False

        path = f'/home/{self.user}/.ssh/known_hosts'
        known = self.target.file.fetch(
            path, mode='600', user=self.user, group=user.group.name
        )
        ok = known.check() is True
        if ok:
            ok = known.hash == self.local.file(path, verbose=False).hash

        return ok

    def deploy(self):
        user = self.target.user.fetch(self.user)

        path = f'/home/{self.user}/.ssh'
        ssh = self.target.dir(
            path, mode='700', user=self.user, group=user.group.name
        )
        if ssh.check() is not True: return False

        full = f'{path}/known_hosts'
        self.target.run('rm -f {full}', sudo='root')
        sent = Sender(self.target, src=full, dst=full, **self.context)()
        if sent is not True: return False

        key = self.target.file(
            full, mode='600', user=self.user, group=user.group.name
        )
        return key.check() is True


class Admin(Actions):
    '''Set User as sudoer
    context:
        user : User name
        pub  : Public key

    - Create User if not present
    - Add User to sudoers
    - Send User's public key
    - Send User's known_hosts
    '''

    @property
    def label(self):
        return f'Admin {self.user}'

    @property
    def actions(self):
        return [
            User_,
            Sudoer,
            CopyId,
            KnownHosts,
        ]




from setux.core.action import Actions, Action


class User_(Action):
    '''Ensure User'''

    @property
    def label(self):
        return 'user'

    def check(self):
        ret, out, err = self.target.run(f'which {self.shell}')
        shell = out[0]
        usr = self.target.user.fetch(self.user,
            uid   = self.uid,
            gid   = self.gid,
            shell = shell,
            home  = f'/home/{self.user}',
        )
        ok = usr.check() is True
        if not ok: return False
        ok = usr.home.check() is True
        return ok

    def deploy(self):
        kw = dict()
        try:
            kw['uid'] = self.uid
        except Exception: pass
        try:
            kw['gid'] = self.gid
        except Exception: pass
        try:
            ret, out, err = self.target.run(f'which {self.shell}')
            kw['shell'] = out[0]
        except Exception: pass
        try:
            kw['home'] = f'/home/{self.user}'
        except Exception: pass
        usr = self.target.user.fetch(self.user, **kw)
        return usr.deploy() is True


class Groups(Action):
    '''Ensure Group'''

    @property
    def label(self):
        return 'groups'

    def check(self):
        grp = self.target.groups.fetch(self.user, *self.groups.split())
        return grp.check() is True

    def deploy(self):
        grp = self.target.groups.fetch(self.user, *self.groups.split())
        return grp.deploy() is True


class User(Actions):
    '''Manage User and its Groups'''

    @property
    def label(self):
        return f'User {self.user}'

    @property
    def actions(self):
        return [
            User_,
            Groups,
        ]




from setux.logger import silent
from setux.core.action import Runner


class Pinger(Runner):
    '''Check target availability'''

    @property
    def labeler(self):
        return silent

    @property
    def label(self):
        return 'ping'

    def deploy(self):
        return self.target.ping()




from setux.core.action import Actions, Action


class hostanme(Action):
    '''Set hostname'''

    @property
    def label(self):
        return 'hostname'

    def check(self):
        ret, out, err = self.target.run('hostname')
        return out[0] == self.hostname

    def deploy(self):
        ret, out, err = self.target.run(f'hostname {self.hostname}', sudo='root')
        if ret: raise RuntimeError('!! ' + err[0])
        return ret == 0


class etc_hostanme(Action):
    '''Set /etc/hostname'''

    @property
    def label(self):
        return '/etc/hostname'

    def check(self):
        current = self.target.read('/etc/hostname').strip()
        return current == self.hostname

    def deploy(self):
        ok = self.target.write(
            '/etc/hostname',
            f'{self.hostname}\n',
            sudo = 'root',
        )
        return ok


class etc_hosts(Action):
    '''Manage /etc/hosts'''

    @property
    def label(self):
        return '/etc/hosts'

    def check(self):
        current = self.target.read('/etc/hosts')
        return self.hostname in current

    def deploy(self):
        current = self.target.read('/etc/hostname').strip()
        new_hosts = self.target.read('/etc/hosts')
        new_hosts = new_hosts.replace(current, self.hostname)
        ok = self.target.write(
            '/etc/hosts',
            new_hosts,
            sudo = 'root',
        )
        return ok


class Hostname(Actions):
    '''Manage Host's Name'''

    @property
    def label(self):
        return f'hostname {self.hostname}'

    @property
    def actions(self):
        return [
            etc_hosts,
            etc_hostanme,
            hostanme,
        ]




from setux.core.action import Runner, Actions


class Moduler(Runner):
    '''Deploy a Module'''

    @property
    def label(self):
        return self.module

    def deploy(self):
        return self.target.deploy(self.module)


class Modules(Actions):
    '''Deploy Modules'''

    @property
    def label(self):
        return f'Modules {self.name}'

    @property
    def actions(self):
        return [
            Moduler(self.target, module=module)
            for module in self.modules
        ]




from setux.logger import silent
from setux.core.action import Action


class Updater(Action):
    '''Update a file'''

    @property
    def label(self):
        return f'upd {self.path}'

    def check(self):
        cont = self.target.read(
            self.path,
            sudo = self.sudo,
            report = 'quiet',
        )
        line = self.line or self.select
        return line in cont

    def deploy(self):
        self.target.deploy('upd_cfg',
            path   = self.path,
            line   = self.line,
            select = self.select,
            remove = self.remove,
            sudo   = self.sudo,
            report = self.report,
        )

        self.target.file(
            self.path,
            user  = self.user,
            group = self.group,
            mode  = self.mode,
            sudo  = self.sudo,
            verbose = False,
        )

        return True



from setux.core.action import Action


class Directory(Action):
    '''Ensure Directory'''

    @property
    def label(self):
        return f'dir {self.path}'

    def check(self):
        directory = self.target.dir.fetch(
            self.path,
            **self.spec,
        )
        return directory.check() is True

    def deploy(self):
        directory = self.target.dir.fetch(
            self.path,
            **self.spec,
        )
        return directory.deploy() is True
